/* FIR_CONVOLUTION ENGINE, VERSION 1.01
  LADSPA INTERFACE TO AN FIR FILTER ENGINE USING TMPFS-BASED I/O
  written by: Charlie Laub, 2020

  For information about installation and usage, please see the instruction file.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  CREDITS:
  I would like to thank Dan Cyr for helping me understand and implement FIR convolution with FFTW.
*/

//FFT_DATATYPE sets the data type used by FFTW
//   double precision data type --> #define FFT_DATATYPE 2
//   default: single precision float type --> #define FFT_DATATYPE 1
#ifndef FFT_DATATYPE
  #define FFT_DATATYPE 1
#endif

const unsigned int NO_DATA_TIMEOUT = 60;
const unsigned int NO_MESSAGE_TIMEOUT = 2;

#include <fstream>
#include <cstdlib>
#include <ctime>
#include <string>
#include <iostream>
#include <chrono>
#include <thread>
#include <cmath>
#include <deque>
#include <vector>
#include <complex>
#include <algorithm>


//needed for FIFOs
#include <string.h> 
#include <fcntl.h> 
#include <sys/stat.h> 
#include <sys/types.h> 
#include <unistd.h> 
#include <signal.h>
#include <stdlib.h>


//needed for FFTW
#include <complex.h>
#include <fftw3.h>


using namespace std;
using namespace std::chrono;




class RunningStat {
  public:
    RunningStat() : m_n(0) {}
    void Clear() {
      m_n = 0;
    }
    void Push(double x) {
      m_n++;
      // See Knuth TAOCP vol 2, 3rd edition, page 232
      if (m_n == 1) {
        m_oldM = m_newM = x;
        m_oldS = 0.0;
      } else {
        m_newM = m_oldM + (x - m_oldM)/m_n;
        m_newS = m_oldS + (x - m_oldM)*(x - m_newM);
        // set up for next iteration
        m_oldM = m_newM; 
        m_oldS = m_newS;
      }
    }
    int NumDataValues() const {
      return m_n;
    }
    double Mean() const {
      return (m_n > 0) ? m_newM : 0.0;
    }
    double Variance() const {
      return ( (m_n > 1) ? m_newS/(m_n - 1) : 0.0 );
    }
    double StandardDeviation() const {
      return sqrt( Variance() );
    }
  private:
    int m_n;
    double m_oldM, m_newM, m_oldS, m_newS;
};


int put_processed_data ( int fd, float * data_container, size_t num_floats ) {
  signal(SIGPIPE, SIG_IGN); //ignore sigpipe errors
  int ret_val;
  //put num_floats*sizeof(float) bytes into FIFO from the data_container
  ret_val = write(fd, data_container, num_floats*sizeof(float) );
  return ret_val;
}


int get_unprocessed_data ( int fd, float * data_container_ptr, size_t num_bytes ) {
  int ret_val;
  //read num_floats*sizeof(float) bytes into data_container from the FIFO
  ret_val = read(fd, data_container_ptr, num_bytes );
  //messages of size >0 contain data. ret_val = -1 indicates empty FIFO
  return ret_val;
} //end function get_unprocessed_data


void send_ready_signal ( int fd ) {
  //By sending one float over the processed_data FIFO the plugin is signaled
  //  that the FIR engine is ready to receive unprocessed data
  signal(SIGPIPE, SIG_IGN); //ignore sigpipe errors
  int ret_val;
  const char * char_array = "READY";
  ret_val = write(fd, char_array, strlen( char_array ));
}
 

bool wait_for_unprocessed_data ( int fd,
                                 unsigned long update_interval,
                                 size_t data_length, 
                                 vector< float > &data_container ) {
  //repeatedly attempt to read the FIFO to see if any data has been delivered
  // returns true after data_length data has been received.
  // if an insufficient amount of data is received after a long time, returns false
 
  auto timer_begin = std::chrono::high_resolution_clock::now();
  std::chrono::seconds elapsed_time;
  unsigned long int initial_interval, wait_interval;
  int ret_val;
  size_t bytes_to_read, samples_read = 0;
  bytes_to_read = 4 * data_length;
  //if update_interval has not yet been set, use a 500 usec wait interval
  //the interval is specified in microseconds
  if ( update_interval == 0 ) {
    wait_interval = 100; 
  } else {
    wait_interval = update_interval / 20;
  }
  initial_interval = wait_interval;
  while ( 1 ) {
    //Attempt to get new unprocessed data from the unprocessed data FIFO. Return values may be:
    //   >0: the number of bytes read
    //    0: no WRITER connected
    //   -1: writer connected but FIFO is empty
    ret_val = get_unprocessed_data( fd, &data_container[samples_read], bytes_to_read );
    
    if ( ret_val > 0 ) {
      //some data was received. Update read counters and return if data_length samples have been obtained 
      bytes_to_read -= ret_val;
      samples_read += ret_val/4;
      if ( bytes_to_read <= 0) return true;
      wait_interval = initial_interval;
    } else {
      // no data received or writer disconnected...
      auto timer_end = std::chrono::high_resolution_clock::now();
      //calculate elapsed time and cast to seconds
      elapsed_time = duration_cast<seconds>(timer_end - timer_begin);
      if ( elapsed_time > std::chrono::seconds( NO_DATA_TIMEOUT ) ) {
        //elapsed time exceeds timeout time, return with false
        return false;
      }
			//double the wait_interval
			wait_interval *= 2;
    } //end if..else
    //timer < 60 second timeout, sleep for 1 wait_interval then check for data again
    this_thread::sleep_for( std::chrono::microseconds( wait_interval ) );
  } //end while loop
} //end function wait_for_unprocessed_data 


std::string& ltrim(std::string& s) {
	auto it = std::find_if(s.begin(), s.end(),
							[](char c) {
								return !std::isspace<char>(c, std::locale::classic());
							});
	s.erase(s.begin(), it);
	return s;
}

std::string& rtrim(std::string& s) {
	auto it = std::find_if(s.rbegin(), s.rend(),
						[](char c) {
							return !std::isspace<char>(c, std::locale::classic());
						});
	s.erase(it.base(), s.end());
	return s;
}

std::string& trim(std::string& s) {
	return ltrim(rtrim(s));
}


void split( deque<string> &res, string s, string delimiter) {
    size_t pos_start = 0, pos_end, delim_len = delimiter.length();
    string token;
    while ((pos_end = s.find (delimiter, pos_start)) != string::npos) {
        token = s.substr (pos_start, pos_end - pos_start);
        pos_start = pos_end + delim_len;
        if (token != "" ) {
          trim( token );
          res.push_back (token);
        }
    }
    token = s.substr (pos_start);
    trim( token );
    res.push_back ( token );
} //end function split



int read_from_control_FIFO ( int fd, deque<string> &messages ) {
  int ret_val;
  char char_array[256];

  ret_val = read(fd, char_array, sizeof(char_array) );
  //messages of size 1 are empty, >1 contain text. ret_val < 1 indicates an error
  if (ret_val > 1 ) {
    string text, delimiter = "|";
    for ( int i=0; i<ret_val; i++ ) {
      text += char_array[i];
    }
    split( messages, text, delimiter );
  }
  return ret_val;
} //end function read_from_control_FIFO 



bool wait_for_messages( unsigned long update_interval, int fd, deque<string> &messages ) {
  //periodically read the FIFO to see if any messages have been delivered
  // returns true after messages have been received.
  // if no messages are received for a long time, returns false
 
  auto timer_begin = std::chrono::high_resolution_clock::now();
  std::chrono::seconds elapsed_time;
  unsigned long int wait_interval;
  int ret_val;
  //if update_interval has not yet been set, use a 1 msec wait interval
  //the interval is specified in microseconds
  if ( update_interval == 0 ) {
    wait_interval = 1000; 
  } else {
    wait_interval = update_interval;
  }
  while ( 1 ) {
    //Check for a new control message:
    //  attempt to read from FIFO and get return value
    //  on success, the value returned is the number of bytes read
    //    0: no WRITER connected
    //    -1: writer connected but FIFO is empty
    ret_val = read_from_control_FIFO ( fd, messages );
    
    if ( ret_val > 0 ) {
      // got a message. Return true
      return true;
    } else {
      // no message received...
      auto timer_end = std::chrono::high_resolution_clock::now();
      //calculate elapsed time and cast to seconds
      elapsed_time = duration_cast<seconds>(timer_end - timer_begin);
      if ( elapsed_time > std::chrono::seconds( NO_MESSAGE_TIMEOUT ) ) {
        //elapsed time exceeds timeout time, return with false
        return false;
      } else {
        //timer < timeout, sleep for 1 wait_interval
        this_thread::sleep_for( std::chrono::microseconds(wait_interval) );
      }
    } //end if..else
  } //end while loop
} //end function wait_for_new_message 



bool process_control_messages ( deque<string> &FIFO_message_queue,
                                deque<string> &saved_messages,
                                string &FIR_filter_filename,
                                string &file_extension,
                                unsigned long &filter_length,
                                unsigned long &data_length,
                                unsigned long &FFT_length,
                                unsigned long &cycle_latency,
                                unsigned long &update_interval,
                                bool &doing_filter_change ) {
  string one_line, keyword, value;
  deque <string> tokens;
  // Process a message into tokens and set program variables
  one_line = FIFO_message_queue[0]; //get the next message from the queue
  split( tokens, one_line, "="); //split the message into tokens
  //there was only one token. The message is not in the keyword = value format, so discard it  
  if ( tokens.size() < 2 ) {
    deque< string >().swap( tokens );
    FIFO_message_queue.pop_front(); //eject the message from the message queue
    return true; //done with this message, skip to end of do..while loop
  } 
  
  // Received a valid message, so process the information
  keyword = tokens[0];
  value = tokens[1];

  if ( keyword == "action" ) {
    if ( value == "terminate" ) {
      return false;
    }
    //save other action messages until initialization has been finished
    saved_messages.push_back( one_line );
    //remove the saved message from the message queue and return
    FIFO_message_queue.pop_front(); 
    return true;
  }   
  if ( keyword == "filter_file" ){
      FIR_filter_filename = value;
      size_t delimiter_position = value.find_last_of(".");
      file_extension = value.substr(delimiter_position+1);
  }
  if ( keyword == "filter_length" ) {
      filter_length = strtoul( value.c_str(),nullptr,0 );
  }
  if ( keyword == "data_length") {
      data_length = strtoul( value.c_str(),nullptr,0 );
  }
  if ( keyword == "FFT_length") {
      FFT_length = strtoul( value.c_str(),nullptr,0 );
  }
  if ( keyword == "cycle_latency") {
      //number of samples that can be accommodated in each I/O file
      cycle_latency = strtoul( value.c_str(),nullptr,0 );
  }
  if ( keyword == "update_interval") {
      //update interval = samples per frame / samples per second, given in microseconds.
      update_interval = stof( value );
  }
  if ( value == "filter_change" ) {
    //set the doing_filter_change flag
    doing_filter_change = true;
  }
  //remove the message from the message queue
  FIFO_message_queue.pop_front();
  return true;
} //end function process_control_messages



int set_FIFO_buffer_size( int fd, size_t capacity, string &message ) {
    //set FIFO buffer capacity in bytes. check, then set, and check again
    long pipe_size = (long)fcntl(fd, F_GETPIPE_SZ);
    if (pipe_size == -1) {
      message = "  could not get FIFO size!";
			return 1;
    }
    int ret = fcntl(fd, F_SETPIPE_SZ, capacity);
    if (ret < 0) {
      message = "  could not set new FIFO size!";
			return 1;
    }
    pipe_size = (long)fcntl(fd, F_GETPIPE_SZ);
    if (pipe_size == -1) {
			message = "  could not confirm new FIFO size.";
			return 1;
    }
    return 0;
}



void cleanup_and_exit (
  ofstream &engine_stats_file,
  string working_directory ) {
  //This function performs cleanup tasks and then exits
  // Close files that are open for I/O
  if ( engine_stats_file.is_open() ) engine_stats_file.close();

	//delete working directory and all files in it using rm -rf <path>
  if ( working_directory != "" ) {
    //use system call to remove directory and files within it
    string rmdir_cmd =  "rm -rf " + working_directory;
    int ret_val;
    ret_val = system( rmdir_cmd.c_str() );
  }

	//terminate FIR engine
  exit(0);
} //end function cleanup_and_exit



//Compile-time error checking of the FFT_DATATYPE
#ifndef FFT_DATATYPE
  #error The FFT_DATATYPE must be defined at the top of the code and assigned the value 1 or 2. 
#endif
#if FFT_DATATYPE != 2
  #if FFT_DATATYPE != 1
    #error ERROR: the FFT_DATATYPE value must be either 1 or 2!
  #endif
#endif


/* ############################################################################
   -------------------------- BEGIN MAIN PROGRAM -----------------------------
   ######################################################################### */


int main( int argc, char* argv[] ) {


  string working_directory;
  //Make sure that working directory filepath was supplied as first command line parameter
  if (argc > 1) {  
    working_directory = argv[1];
  } else {
    //Fatal error: the program was not called with at least one parameter. Exit
    exit(1);
  }
  //make sure that directory actually exists...
  int i;
  string temp_string;
  temp_string = "ls " + working_directory + " > /dev/null 2>&1";
  i = system( temp_string.c_str() );
  if ( i != 0 ) {
    //Fatal error: the supplied working directory does not exit. Exit
    exit(2);
  }

  RunningStat rs; //declare new running stats instance

  //timing related vars
  std::chrono::microseconds longest_cycle_time = std::chrono::microseconds::zero();
  std::chrono::microseconds cycle_time;
  std::chrono::seconds elapsed_time;

  ifstream config_file, FIR_filter_file;
  ofstream engine_stats_file, error_file;
  unsigned long int data_length, filter_length, FFT_length, cycle_latency, update_interval;
  int ret_val, control_fifo_fd, unprocessed_data_fd, processed_data_fd;
  unsigned int line_number;
  string one_line, FIR_filter_filename, file_extension;
  float normalization_factor;
  bool doing_filter_change = false;
	bool is_first_cycle = true; //this flag is set to false after the first convolution has been performed
  bool doing_initial_startup = true; //this flag is set to false at the end of the initial startup code

  //STL containers:
  deque< string > FIFO_message_queue, saved_messages, tokens;
  vector< float > unprocessed_data;
  vector< float > processed_data;
  vector< float > FIR_filter;

  //set the data type based on the value of FFT_DATATYPE
  #if FFT_DATATYPE == 2
    typedef double FFT_data_type;
  #else
    typedef float FFT_data_type;
  #endif

  //declare the STL containers that will hold real valued data
  vector< FFT_data_type > overlap_data;

  //declare the STL containers that will hold complex data
  vector< complex< FFT_data_type > > outfftwOne;
  vector< complex< FFT_data_type > > outfftwTwo;
  vector< complex< FFT_data_type > > fftMulti;
  

  //declare plans based on the FFT_DATATYPE:
  #if FFT_DATATYPE == 2      
  	//Declare double precision forward FFT plans
  	fftw_plan audio_fFFT;
  	fftw_plan filter_fFFT;
  	//Declare double precision backward (inverse) FFT plan
  	fftw_plan filterXaudio_iFFT;
  #else
  	//Declare single precision forward FFT plans
  	fftwf_plan audio_fFFT;
  	fftwf_plan filter_fFFT;
  	//Declare single precision backward (inverse) FFT plan
  	fftwf_plan filterXaudio_iFFT;
  #endif
    

  //FILE to which errors are reported
  temp_string = "/dev/shm/FIR_engine.log";  
  error_file.open( temp_string.c_str(), ios_base::app );


  //Initialize FIFOs:
  //NOTE: the FIFO files must have already been created in the OS before calling this program
  
  //Create and open control FIFO read-only , non-blocking. Retuns with file descriptor
  temp_string = working_directory + "/control"; 
  control_fifo_fd = open(temp_string.c_str(), O_RDONLY | O_NONBLOCK);
  //now ready to receive parameters and action commands over the FIFO.

  //open processed_data FIFO as writer in blocking mode. A write will block if the FIFO is full
  temp_string = working_directory + "/processed_data"; 
  processed_data_fd = open(temp_string.c_str(), O_WRONLY );

  //open unprocessed_data FIFO read-only , non-blocking. Retuns with file descriptor
  temp_string = working_directory + "/unprocessed_data"; 
  unprocessed_data_fd = open(temp_string.c_str(), O_RDONLY | O_NONBLOCK);

  //set FIFO capacity in bytes. 
  // NOTE: 1048576 bytes = 1 MiB is the largest allowed by the kernel
  temp_string = "";
  ret_val = set_FIFO_buffer_size( unprocessed_data_fd, 1048576, temp_string );
  if ( temp_string != "" ) error_file << temp_string << endl;
  // if function returns 1 an error occurred, so terminate
  if ( ret_val != 0 ) cleanup_and_exit( engine_stats_file, working_directory );

  //open the cycle time information file 
  if ( ! engine_stats_file.is_open() ) {
    temp_string = working_directory + "/cycle_time_stats";  
    engine_stats_file.open( temp_string.c_str() );
  }
    



	//initialize critical FIR engine parameters
  FIR_filter_filename = "";
  file_extension = "";
	filter_length = 0;
	data_length = 0;
	FFT_length = 0;
	cycle_latency = 0;
	update_interval = 0;

  unsigned long longest_cycle_number, cycle_counter = 0;


  //begin main loop
  while ( 1 ) {
    //Get FIR engine parameters from the LADSPA plugin via FIFO
    do { 
      // This do..while loop runs until all critical params have been received
      if ( FIFO_message_queue.size() == 0 ) {
        //the message queue is empty. Waiting until new messages arrive
        if ( ! wait_for_messages( update_interval, control_fifo_fd, FIFO_message_queue ) ) {
          // if function returns false, too much time has elapsed, so terminate
          cleanup_and_exit( engine_stats_file, working_directory );
        }
      }
      // Received one or more messages, so process them
      if ( ! process_control_messages( FIFO_message_queue,
                                saved_messages,
                                FIR_filter_filename,
                                file_extension,
                                filter_length,
                                data_length,
                                FFT_length,
                                cycle_latency,
                                update_interval,
                                doing_filter_change ) ) {
        //process_control_messages returned false. This means that action = terminate was received,
        //  so do cleanup and terminate
        cleanup_and_exit( engine_stats_file, working_directory );
      }

    } while (
          ( FIR_filter_filename == "" ) ||
          ( filter_length == 0 ) ||
          ( data_length == 0 ) ||
          ( FFT_length == 0 ) ||
          ( cycle_latency == 0 )
          ); //end do loop that waits for all critical FIR engine parameters

    // Critical FIR engine parameters have been received. Proceed with initialization...
    
    //computer normalization_factor
    normalization_factor = static_cast< FFT_data_type > ( FFT_length );
    normalization_factor = 1.0 / normalization_factor;

    //create/resize storage for the FIR filter before populating with filter data
    //   and initialize contents to zeros
    if ( FIR_filter.size() < FFT_length ) FIR_filter.resize( FFT_length, 0.0 );
  
    //Read in the FIR filter from file
    //open file
    FIR_filter_file.open(FIR_filter_filename.c_str());
    //reading of file depends on file type:
    if ( file_extension == "txt" ) {
      //text file: one tap per line, initial comment lines possible
      //getline into string; stof converts it to a float
      line_number = 0;
      //first we check for comment lines
      while ( getline( FIR_filter_file, one_line ) ) {
        if ( one_line.find("#") != string::npos ) {
          //skip comment lines indicated by "#"
          continue;
        } else {
          //first data is found, put it in the filter
          FIR_filter[ line_number ] = stof( one_line );          
          line_number++;
          break;
        }
      }
      //no need to check for comment lines after first data is found...
      //to prevent array from overwriting its bounds, check against filter length
      while ( getline( FIR_filter_file, one_line ) && ( line_number < filter_length ) ) {
        FIR_filter[ line_number ] = stof( one_line );
        line_number++;
      }
    }
    if ( file_extension == "bin" ) {
      //read in binary data file (floats)
      FIR_filter_file.read( reinterpret_cast<char *>(&FIR_filter[0]), filter_length*sizeof(float) );
    }
    if ( file_extension == "csv" ) {                                //NEED TO REDO CSV READ WITH MORE EFFICIENT METHOD..
      //read in comma separated values data file (float or doubles)
      // idea: read each char and buffer input, when comma is reached convert buffer to float
      char one_char;
      string str_buffer;
      line_number = 0;
      while ( FIR_filter_file.get(one_char) && ( line_number < filter_length ) ) {
        if ( one_char == ',' ) {
          FIR_filter[ line_number ] = stof( str_buffer );
          str_buffer = "";
          line_number++;
        } else {
          str_buffer += one_char;
        }
      }
    }
    //done reading FIR filter file, so close it.
    FIR_filter_file.close();


/**********************************************************************************
                               set up the convolution
**********************************************************************************/

    //if this is during startup, perform FFTW related initializations
    if ( doing_initial_startup ) {  
    
      //size the STL containers that will hold real valued data and fill with zeros
      overlap_data.resize( filter_length-1, 0.0 );       
      unprocessed_data.resize( FFT_length );
      FIR_filter.resize( FFT_length );
      processed_data.resize( FFT_length );
    
      //size the STL containers that will hold complex data
       outfftwOne.resize( FFT_length );
       outfftwTwo.resize( FFT_length );
       fftMulti.resize( FFT_length/2 + 1 ); //num_complex_data = FFT_length/2 + 1

// NOTE OCT2022: wisdom not currently used, but this retained for possible future use    	
//       //Import FFTW system wisdom:
//       #if FFT_DATATYPE == 2      
//         //get FFTW System Wisdom from /etc/fftw/wisdom
//         ret_val = fftw_import_system_wisdom(); //import double precision wisdom
//       #else
//         //get FFTW System Wisdom from /etc/fftw/wisdomf
//         ret_val = fftwf_import_system_wisdom(); //import single precision wisdom
//       #endif
//       if ( ret_val != 1 ) {
// 				//ret_val is 1 on successful wisdom import
//         error_file << "An error ocurred when trying to import system Wisdom." << endl;
//         cleanup_and_exit( engine_stats_file, working_directory );
//       }

      #if FFT_DATATYPE == 2      
      	//create the FFT plans in double precision
      	audio_fFFT = fftw_plan_dft_r2c_1d(
            FFT_length, 
            &unprocessed_data[0],
            reinterpret_cast<fftw_complex*>(&outfftwOne[0]), 
            FFTW_ESTIMATE );
      	filter_fFFT = fftw_plan_dft_r2c_1d(
            FFT_length, 
            &FIR_filter[0], 
            reinterpret_cast<fftw_complex*>(&outfftwTwo[0]), 
            FFTW_ESTIMATE );
      	filterXaudio_iFFT = fftw_plan_dft_c2r_1d(
            FFT_length, 
            reinterpret_cast<fftw_complex*>(&fftMulti.[0]), 
            &processed_data[0], 
            FFTW_ESTIMATE );
      #else
      	//create the FFT plans in single precision
        audio_fFFT = fftwf_plan_dft_r2c_1d(
            FFT_length, 
            &unprocessed_data[0],
            reinterpret_cast<fftwf_complex*>(&outfftwOne[0]), 
            FFTW_ESTIMATE );
      	filter_fFFT = fftwf_plan_dft_r2c_1d(
            FFT_length, 
            &FIR_filter[0], 
            reinterpret_cast<fftwf_complex*>(&outfftwTwo[0]), 
            FFTW_ESTIMATE );
      	filterXaudio_iFFT = fftwf_plan_dft_c2r_1d(
            FFT_length, 
            reinterpret_cast<fftwf_complex*>(&fftMulti[0]), 
            &processed_data[0], 
            FFTW_ESTIMATE );
      #endif
      
      //make sure all FFTW plans were created sucessfully
      if (  (filter_fFFT == NULL) || (filter_fFFT == 0)    ||
            (audio_fFFT == NULL)  || (audio_fFFT == 0)     ||
            (filterXaudio_iFFT == NULL) || (filterXaudio_iFFT == 0) 
          ) { 
      	error_file << "One or more FFTW plans could not be created!" << endl;
        cleanup_and_exit( engine_stats_file, working_directory );
      }
    } //end if ( doing_initial_startup ) FFTW related initializations


    //perform filter related fFFT, done only at startup or upon filter change
    if ( ( doing_filter_change ) || ( doing_initial_startup ) ) {
    	//zero-pad the filter only if filter change
      if ( doing_filter_change ) {
      	for (i = filter_length; i < FFT_length; i++) FIR_filter[i] = 0.0;
      } 
      //compute the forward transform of the filter
      #if FFT_DATATYPE == 2  
        //compute fFFT of the filter in double precision
      	fftw_execute(filter_fFFT);
      #else
        //compute fFFT of the filter in single precision
      	fftwf_execute(filter_fFFT);
      #endif
    } //end if doing_filter_change or is_first_cycle   

  

    //If the initializations came about in response to a filter change, run the convolution again now
    //  The unprocessed data should not have been altered by the function previously because 
    //  the convolution is not inplace, therefore, the new FIR filter can be run on  
    //  the 'old' data' to generate a new set of overlap-add data for the next cycle
    if ( doing_filter_change ) {
      // THIS CODE ONLY NEEDS TO GENERATE NEW OVERLAP DATA
      i = 0; //A DUMMY STATEMENT FOR NOW...
      //THIS IS NOT YET IMPLEMENTED.
    }
    //clear the doing_filter_change flag
    doing_filter_change = false;

  
    // Initialize time variables to current time
    auto timer_begin = std::chrono::high_resolution_clock::now();
    auto timer_end = std::chrono::high_resolution_clock::now();
    auto last_stats_write = std::chrono::high_resolution_clock::now();


    if ( doing_initial_startup ) {
      send_ready_signal( processed_data_fd );
      doing_initial_startup = false;
    }

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                                 Done With Initialization.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */ 
  
    //begin loop containing the FIR filter convolution
    do {

      //check for new messages in the control FIFO, or or existing messages, and process them
      if (    ( read_from_control_FIFO ( control_fifo_fd, FIFO_message_queue ) > 0 )
           || ( FIFO_message_queue.size() > 0 ) 
         ) process_control_messages( 
              FIFO_message_queue,
              saved_messages,
              FIR_filter_filename,
              file_extension,
              filter_length,
              data_length,
              FFT_length,
              cycle_latency,
              update_interval,
              doing_filter_change );


      //Get a new set of data from the unprocessed_data FIFO. 
      if ( ! wait_for_unprocessed_data ( unprocessed_data_fd, update_interval, data_length, unprocessed_data ) ) {
        // if function returns false, too much time has elapsed, so terminate
        cleanup_and_exit( engine_stats_file, working_directory );
      }

      //set timer_begin to current time. Used to determine convolution cycle time
      timer_begin = std::chrono::high_resolution_clock::now();

    /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                                     Perform the convolution:
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */ 
  
    	//copy unprocessed_data and zero-pad it to generate sigOnePadded 
    	for (i = data_length; i < FFT_length; i++) unprocessed_data[i] = 0.0;
 
      #if FFT_DATATYPE == 2      
        //compute fFFT of sigOnePadded in double precision
      	fftw_execute(audio_fFFT);
      #else
        //compute fFFT of sigOnePadded in single precision
      	fftwf_execute(audio_fFFT);
      #endif    

      //Complex multiplication: fFFT(data)*fFFT(filter)
    	for ( i = 0; i < (FFT_length/2 + 1); i++ ) fftMulti[i] = outfftwOne[i] * outfftwTwo[i]; 
    
      #if FFT_DATATYPE == 2      
        //compute iFFT of fFFT(data)*fFFT(filter) in double precision
      	fftw_execute(filterXaudio_iFFT);
      #else      
        //compute iFFT of fFFT(data)*fFFT(filter) in single precision
      	fftwf_execute(filterXaudio_iFFT);
      #endif      

      //normalize the output
      for (i=0; i<FFT_length; i++) processed_data[i] *= normalization_factor;
      
      //perform overlap-add using saved data from last call 
      for (i=0; i<(filter_length-1); i++) processed_data[i] += overlap_data[i];
      //save new overlap data
      //the overlap data is located between indeces data_length and data_length+filter_length-1
      for (i=0; i<(filter_length-1); i++) overlap_data[i] = processed_data[data_length+i];

  		cycle_counter++; //increment the convolution cycle counter

      //write data_length samples to processed_data FIFO  
      //processed_data FIFO should block until write can be conpleted...
      ret_val = put_processed_data( processed_data_fd, &processed_data[0], data_length );
    
      // Done with FIR convolution and I/O. Get system time and compute elapsed cycle time
      timer_end = std::chrono::high_resolution_clock::now();
      //calculate elapsed time and cast to microseconds
      cycle_time = duration_cast<microseconds>(timer_end - timer_begin);
      //if this is the longest time ever, replace the previous value
      if ( cycle_time > longest_cycle_time ) {
				longest_cycle_time = cycle_time;
				longest_cycle_number = cycle_counter;
			}

      //if this is the first cycle, sleep until cycle_latency has elapsed. 
      //this provides buffering against varying FIR engine cycle times.
      if ( is_first_cycle ) {
        if ( (cycle_latency - cycle_time.count()) > 0 ) 
          this_thread::sleep_for( std::chrono::microseconds( (cycle_latency - cycle_time.count()) ) );
        is_first_cycle = false;
      }

      //only write stats to file if we will NOT be doing a filter change.
      if ( ! doing_filter_change ) {    
        // push cycle time in usec into running stats
        rs.Push( (double)cycle_time.count() );
        // Write running stats to file in working directory every 2 minutes
        //check times elapsed since last write  
        elapsed_time = duration_cast<seconds>(timer_end - last_stats_write);
        if ( elapsed_time > std::chrono::seconds(10) ) {
          //rewind to beginning of file to overwrite the previous data
          engine_stats_file.seekp(0);
          //write stats to file  
          engine_stats_file << "Cycle time running statistics (times in usec):\n";
          engine_stats_file << "  mean = " << rs.Mean() << "\n";
          engine_stats_file << "  standard deviation of mean = " << rs.StandardDeviation() << "\n";
          engine_stats_file << "longest ever cycle time = " << longest_cycle_time.count() << "\n";
					engine_stats_file << "   occured during cycle " << longest_cycle_number << endl;
          //update last_stats_write time
          last_stats_write = std::chrono::high_resolution_clock::now();
        }
      } //done writing stats
    } while ( doing_filter_change == false ); //End loop containing the FIR filter convolution
    //only get here if the code is changing over to a new FIR filter
    //reset stats related data
    rs.Clear(); 
    //clear FIR_filter_filename in anticipation of a new filter
    FIR_filter_filename == "";
  } //end main loop
} //end program