/* LADSPA_FIR PLUGIN, VERSION 1.0 
  LADSPA INTERFACE TO AN FIR FILTER ENGINE USING TMPFS-BASED I/O
  written by: Charlie Laub, 2020

  For information about installation and usage, please see the instruction file.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  CREDITS:
  I would like to thank Dan Cyr for helping me understand and implement FIR convolution with FFTW.
*/          

#include <iostream>
#include <fstream>
#include <cstdlib>
#include <ctime>
#include <string>
#include <deque>
#include <vector>
#include <algorithm>
#include <chrono>

#include <ladspa.h>
#include <stdlib.h>
#include <math.h>

//these c-style headers needed for FIFOs!
#include <stdio.h> 
#include <string.h> 
#include <fcntl.h> 
#include <sys/stat.h> 
#include <sys/types.h> 
#include <unistd.h> 
#include <signal.h>

// Linux specific fcntl header to set FIFO size
// #include </usr/include/linux/fcntl.h>

using namespace std;

//LADSPA parameter order:
#define FILTER_INDEX      0
#define DATA_LENGTH       1
#define FFT_LENGTH        2
#define CYCLE_LATENCY     3
#define FILTER_LABEL      4
#define DATA_INPUT        5
#define DATA_OUTPUT       6



//global variables
ofstream error_file;



unsigned long x_to_the_p(unsigned long x, unsigned long p) {
  //exponentiation function for integers: x^p
  if (p == 0) return 1;
  if (p == 1) return x;
  unsigned long tmp = x_to_the_p(x, p/2);
  if (p%2 == 0) return tmp * tmp;
  else return x * tmp * tmp;
} //end function x_to_the_p


unsigned long get_factors(unsigned long n, vector< unsigned long > &factors ) {
  //factor a number by the primes 2, 3, 5, and 7. Return the remainder.
  unsigned long z = 2;
  while (z * z <= n) {
    if (n % z == 0) {
      factors[z] += 1;
      n /= z;
    } else {
      z++;
      if ( z > 7 ) return n;
    }
  }
  return n; 
} //end function get_factors


unsigned long optimize_for_FFTW_DFT( unsigned long trial_num ) {
  //check to see if trial_num is a factor of only the prime numbers 2, 3, 5, and 7
  //  if it is, return trial_num
  //  if not, find the next highest number that is and return that
  unsigned long new_num, num_to_factor;
  unsigned long lowest_best = 0;
  vector< unsigned long > factors(8,0);
  
//  cout << "OPTIMIZING INPUT VALUE OF :" << trial_num << " FOR FFTW." << endl;
  do {
    num_to_factor = trial_num;
    factors[2] = 0;
    factors[3] = 0;
    factors[5] = 0;
    factors[7] = 0;
  
    new_num = get_factors(  num_to_factor, factors );
    while (new_num > 7 ) {
      num_to_factor = new_num + 1;
      new_num = get_factors(  num_to_factor, factors );
    } 
    factors[new_num] += 1;
  
    new_num = x_to_the_p( 2, factors[2] );
    new_num *= x_to_the_p( 3, factors[3] );
    new_num *= x_to_the_p( 5, factors[5] );
    new_num *= x_to_the_p( 7, factors[7] );
  
    if ( lowest_best == 0 ) lowest_best = new_num;
    if ( new_num < lowest_best ) lowest_best = new_num;
  
    trial_num += 1;
  
  } while ( lowest_best > trial_num );
  
//   cout << "THE NEXT NUMBER OF THE FORM 2^a * 3^b * 5^c * 7^d" << endl;
//   cout << "   2 ^ " << factors[2] << " * ";
//   cout << "3 ^ " << factors[3] << " * ";
//   cout << "5 ^ " << factors[5] << " * ";
//   cout << "7 ^ " << factors[7] << endl;
//   cout << "IS: " << lowest_best << endl;
  
  return lowest_best;
} //end function optimize_for_FFTW_DFT




//trim whitespace the the left of text
std::string& ltrim(std::string& s)
{
	auto it = std::find_if(s.begin(), s.end(),
		[](char c) {
			return !std::isspace<char>(c, std::locale::classic());
		});
	s.erase(s.begin(), it);
	return s;
}

//trim whitespace the the right of text
std::string& rtrim(std::string& s)
{
	auto it = std::find_if(s.rbegin(), s.rend(),
		[](char c) {
			return !std::isspace<char>(c, std::locale::classic());
		});
	s.erase(it.base(), s.end());
	return s;
}

//trim whitespace the the left and right of text
std::string& trim(std::string& s)
{
	return ltrim(rtrim(s));
}


void split( deque<string> &res, string s, string delimiter) {
    size_t pos_start = 0, pos_end, delim_len = delimiter.length();
    string token;
    while ((pos_end = s.find (delimiter, pos_start)) != string::npos) {
        token = s.substr (pos_start, pos_end - pos_start);
        pos_start = pos_end + delim_len;
        if (token != "" ) {
          trim( token );
          res.push_back (token);
        }
    }
    token = s.substr (pos_start);
    trim( token );
    res.push_back ( token );
}//end function split




int get_processed_data ( int fd, float * data_container, size_t num_floats ) {
  int ret_val;

  //read num_floats*sizeof(float) bytes into data_container from the FIFO
  ret_val = read(fd, data_container, num_floats*sizeof(float) );
  //messages of size >0 contain data. ret_val = -1 indicates empty FIFO
  return ret_val;
} //end function read_from_FIFO 


int put_unprocessed_data ( int fd, float * data_container, size_t num_floats ) {
  signal(SIGPIPE, SIG_IGN); //ignore sigpipe errors
  int ret_val;

  //put num_floats*sizeof(float) bytes into FIFO from the data_container
  ret_val = write(fd, data_container, num_floats*sizeof(float) );

  return ret_val;
}


bool check_for_FIR_engine_ready_signal( int fd ) {
  //try to read a short text string from the processed_data FIFO
  //"READY" is sent by the FIR engine to indicate it has completed initialization and is ready
  int ret_val;
  char char_array[16];
  ret_val = read(fd, char_array, sizeof(char_array) );
  //messages of size >0 contain data. ret_val = -1 indicates empty FIFO
  if ( ret_val > 0 ) return true;
  return false;
}


void create_FIFO_file( string fifo_filename ) {
	const char * myfifo = fifo_filename.c_str(); 
	// Create the named file(FIFO) 
	// mkfifo(<pathname>, <permission>) 
	mkfifo(myfifo, 0666); 
}


void queue_message_for_FIFO( string &one_message, string &FIFO_messages ) {
  if ( FIFO_messages != "" ) FIFO_messages += "|";
  FIFO_messages += one_message;
  one_message = "";
}

                                                  
void write_to_control_FIFO ( int fd, string &messages, bool &control_FIFO_open ) {
  signal(SIGPIPE, SIG_IGN); //ignore sigpipe errors
  int ret_val;
  const char * char_array = messages.c_str();
  ret_val = write(fd, char_array, strlen( char_array ));
  if (ret_val > 0 ) {
    messages = ""; //messages successfully sent, so empty the variable
  } else {
    control_FIFO_open = false;
    close(fd);
  }
}


void set_FIFO_buffer_size( int fd, size_t capacity ) {
    //set FIFO buffer capacity in bytes. check, then set, and check again
    long pipe_size = (long)fcntl(fd, F_GETPIPE_SZ);
    if (pipe_size == -1) {
      error_file << "  could not get FIFO size!" << endl;
			return;
    }
    int ret = fcntl(fd, F_SETPIPE_SZ, capacity);
    if (ret < 0) {
      error_file << "  could not set new FIFO size!" << endl;
			return;
    }
    pipe_size = (long)fcntl(fd, F_GETPIPE_SZ);
    if (pipe_size == -1) {
			error_file << "  could not confirm new FIFO size." << endl;
			return;
    }
}



string get_random_char_str( const int num_chars ) {
	/* Decleration and initialization of static constant string type variable */
	static bool rand_is_initialized = false;
	static const string charList = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
	if ( ! rand_is_initialized ) {
		/* srand(): initialize random number generator from the current time */
		auto now = std::chrono::system_clock::now();
		auto now_us = std::chrono::time_point_cast<std::chrono::microseconds>(now);
		auto epoch = now_us.time_since_epoch();
		auto epoch_usec = std::chrono::duration_cast<std::chrono::microseconds>(epoch);
		unsigned int seed;
		seed = static_cast<unsigned int>(epoch_usec.count() );
		srand(seed);
		rand_is_initialized = true;
	}

	string alphanumeric;
  //create 8-character-long string of random alphanumeric characters	
	for(int i = 0; i < num_chars; i++) {
		/* rand() generate random number */
		alphanumeric += charList [rand() % charList.size()];
	}
  return alphanumeric;
}


string GetEnv( const std::string & var ) {
  //put the value of an OS environmental variable into a string
  const char * val = std::getenv( var.c_str() );
  if ( val == nullptr ) { // invalid to assign nullptr to std::string
    return "";
  }
  else {
    return val;
  }
}


int run_OS_command( string some_command ) {
  //run string passed as command in the OS
  // Convert string to const char * as system requires 
  const char *command = some_command.c_str();
  int ret_val; 
  ret_val = system(command);
  return ret_val;   
}


size_t getFilesize(const std::string& filename) {
    struct stat st;
    if(stat(filename.c_str(), &st) != 0) {
        return 0;
    }
    return st.st_size;   
}


inline bool file_exists (const std::string& name) {
  struct stat buffer;   
  return (stat (name.c_str(), &buffer) == 0); 
}

static LADSPA_Descriptor *FIR_filter_Descriptor = NULL;


typedef struct {
	LADSPA_Data *input;
	LADSPA_Data *output;
  unsigned long data_length;
  unsigned long filter_length;
  unsigned long FFT_length;
  unsigned long cycle_latency;
  unsigned long SR; //the sample rate
  unsigned long call_counter;
  unsigned long latent_sample_count;
  unsigned int user_filter_ID;
  unsigned int filter_label;
  int control_fifo_fd;
  int unprocessed_data_fd;
  int processed_data_fd;
  char messages[256];
  char working_directory[128];
  bool already_activated;
  bool control_FIFO_open;
  bool fatal_errors_encountered;
  bool FIR_engine_is_ready;
  bool is_startup;
} plugin_data_struct;




const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return FIR_filter_Descriptor;
	default:
		return NULL;
	}
}


LADSPA_Handle instantiate(const LADSPA_Descriptor *descriptor, unsigned long sample_rate) {
  plugin_data_struct *FIR = (plugin_data_struct *)malloc(sizeof(plugin_data_struct));

  //store the sample rate  
  FIR->data_length = 0;
  FIR->filter_length = 0;
  FIR->FFT_length = 0;
  FIR->cycle_latency = 0;
  FIR->SR = sample_rate;
  FIR->call_counter = 0;
  FIR->latent_sample_count = 0;
  FIR->already_activated = false;
  FIR->control_FIFO_open = false;
  FIR->fatal_errors_encountered = false;
  FIR->FIR_engine_is_ready = false;
  FIR->is_startup = true;


  return (LADSPA_Handle)FIR;
}


void connectPort(LADSPA_Handle instance, unsigned long port, LADSPA_Data *data) {
  plugin_data_struct *FIR = (plugin_data_struct *)instance;
  float temp;
  string temp_string;
	switch (port) {
  	case 0:
  		FIR->user_filter_ID = static_cast<unsigned int>( *data );
  		break;
  	case 1:
      temp = *data; 
      temp = abs( trunc( temp ) ); //truncate to whole number, make sure is >= 0       
      FIR->data_length = static_cast<unsigned long>(temp); //cast to long int
  		break;
  	case 2:
      temp = *data; 
      temp = abs( trunc( temp ) ); //truncate to whole number, make sure is >= 0       
      FIR->FFT_length = static_cast<unsigned long>(temp); //cast to long int
  		break;
  	case 3:
      //latency in microseconds, passed as float
      temp = *data; 
      temp = abs( trunc( temp ) ); //truncate to whole number, make sure is >= 0       
      FIR->cycle_latency = static_cast<unsigned long>(temp); //cast to long int
  		break;
  	case 4:
      //filter label, whole number part is used as the dir name for the filter, passed as float
      temp = *data; 
      temp = abs( trunc( temp ) ); //truncate to whole number, make sure is >= 0       
      FIR->filter_label = static_cast<unsigned int>(temp); //cast to u_int;
  		break;
  	case 5:
  		FIR->input = data;
  		break;
  	case 6:
  		FIR->output = data;
  		break;
	} //end switch
} //end function connect_port


void activate(LADSPA_Handle instance) {
  plugin_data_struct *FIR = (plugin_data_struct *)instance;

  if ( ! FIR->already_activated ) {
    //PERFORM ONE-TIME FIR ENGINE START-UP TASKS:
    FIR->already_activated = true; //set to true so this code does not run again.
    string temp_string;
    int ret_val;
    string dir_name, OS_commands;

    //  GET VALUES OF ENVIRONMENTAL VARIABLES FOR FILTER TABLE AND FIR ENGINE
    string FIR_filter_table_filename, FIR_engine_cmdstr;
    FIR_filter_table_filename = GetEnv("LADSPA_FIR_FILTER_TABLE");
    FIR_engine_cmdstr = GetEnv("LADSPA_FIR_FILTER_ENGINE");
    if ( ( FIR_filter_table_filename == "" ) || ( FIR_engine_cmdstr == "" ) ) {
      //At least one critical environmental variable was not found. Set fatal error flag
      FIR->fatal_errors_encountered = true;
      if ( FIR_filter_table_filename == "" )
        error_file << "The environmental variable LADSPA_FIR_FILTER_TABLE was not found. " << endl;
      if ( FIR_engine_cmdstr == "" )
        error_file << "The environmental variable LADSPA_FIR_FILTER_ENGINE was not found. " << endl;
      return;
    } else {
      //check to make sure that the files exist
      temp_string = FIR_filter_table_filename;
      if (! file_exists( temp_string ) ){
        FIR->fatal_errors_encountered = true;
        error_file << "The file " << FIR_filter_table_filename << " was not found. " << endl;
        return;
      }
      temp_string = FIR_engine_cmdstr;
      if (! file_exists( temp_string ) ){
        FIR->fatal_errors_encountered = true;
        error_file << "The file " << FIR_engine_cmdstr << " was not found. " << endl;
        return;
      }
    }

    //  Read the FIR filter table to obtain the number of taps and filter file pathname
    string one_line, filter_file;
    deque< string > tokens;
    unsigned int ID_value;
  
    //attempt to open the FIR filter file and perform error handling
    ifstream FIR_filter_table;
    FIR_filter_table.open( FIR_filter_table_filename.c_str() );
    if ( ! FIR_filter_table ) {
      //an error was thrown when trying to open the filter table. Set fatal error flag
      FIR->fatal_errors_encountered = true;
      //error message text is contained in the string  e.code().message()
      error_file << "When trying to open the filter table file, an error occurred ";
			error_file << "or the file could not be opened." << endl; 
      return;
    }  

    while ( getline( FIR_filter_table, one_line ) ) {
      if ( ( one_line.find("#") != string::npos ) || ( one_line == "" ) ) continue; //skip comment and blank lines
      split( tokens, one_line, " " );
      ID_value = atoi( tokens[0].c_str() );
      if ( ID_value == FIR->user_filter_ID ) break;
      deque< string >().swap( tokens ); //not a match, so discard the tokens
    } //end while
    
    //done reading the FIR filter table so close the file
    FIR_filter_table.close();

    if ( tokens.size() == 3 ) {
      FIR->filter_length = stoi( tokens[1] );
      filter_file = tokens[2];
    } else {
      FIR->fatal_errors_encountered = true;
      //the filter ID was not found in the filter table or the line did not 
      //  contain three tokens. Set fatal error flag    
      error_file << "ERROR: The filter ID was not found in the filter table, or ";
      error_file << "the line did not contain three tokens." << endl;
      return; 
    }
    
    //check that FFT_length >= data_length + filter_length
    if ( ! ( FIR->FFT_length >= (FIR->data_length + FIR->filter_length) ) ) {
      FIR->fatal_errors_encountered = true;
      error_file << "ERROR: The FFT length must be >= data_length + filter_length" << endl;
      return; 
    }
    
    //check that data_length > filter_length
    if ( ! (FIR->data_length >= (2*FIR->filter_length) ) ) {
      FIR->fatal_errors_encountered = true;
      error_file << "ERROR: the data length must be >= twice the filter length" << endl;
      return; 
    }
    
    //optimize the FFT_length so that it is a factor of only the primes 2, 3, 5, and 7
    //  This may result in a slightly larger FFT_length than what was requested by the 
    //  user, however, this maximizes the FFTW speed
    FIR->FFT_length = optimize_for_FFTW_DFT( FIR->FFT_length );

    //	Queue up info to be sent over the control FIFO
    string a_message, all_messages; 
    a_message = "filter_length = " + to_string(FIR->filter_length);
    queue_message_for_FIFO( a_message, all_messages );

    a_message = "filter_file = " + filter_file;
    queue_message_for_FIFO( a_message, all_messages );

    a_message = "data_length = " + to_string(FIR->data_length);
    queue_message_for_FIFO( a_message, all_messages );

    a_message = "FFT_length = " + to_string(FIR->FFT_length);
    queue_message_for_FIFO( a_message, all_messages );

    a_message = "cycle_latency = " + to_string(FIR->cycle_latency);
    queue_message_for_FIFO( a_message, all_messages );

		strcpy(FIR->messages, all_messages.c_str() );


    //create unique working directory in the tmpfs
    if ( FIR->filter_label == 0 ) {
      dir_name = "/dev/shm/FIR_" + get_random_char_str( 6 );
    } else {
      temp_string = to_string( FIR->filter_label );
      //pad the front of the string with zeroes until the string is four characters long
      while ( temp_string.size() < 4 ) {
        temp_string.insert(0,"0");
      }
      dir_name = "/dev/shm/FIR_" + temp_string + "." + get_random_char_str( 2 );
    }
    OS_commands = "mkdir " + dir_name;
    ret_val = run_OS_command( OS_commands );  
    if ( ret_val != 0 ) {
      //an error was thrown when trying to make the unique directory. Set fatal error flag
      FIR->fatal_errors_encountered = true;
      error_file << "An error was thrown when trying to make the unique directory " << dir_name << endl;
      return;
    }
    strcpy( FIR->working_directory, dir_name.c_str() );  //store for later


  	//Create control FIFO file 
    string fifo_filename;
    fifo_filename = dir_name + "/control";  
  	create_FIFO_file( fifo_filename ); 
  	//Create unprocessed_data FIFO file 
    fifo_filename = dir_name + "/unprocessed_data";  
  	create_FIFO_file( fifo_filename ); 
  	//Create processed_data FIFO file 
    fifo_filename = dir_name + "/processed_data";  
  	create_FIFO_file( fifo_filename ); 


    //open the processed_data FIFO read-only , non-blocking. Retuns with file descriptor
    //this will always succeed unless the fifo file cannot be created.
    temp_string = dir_name + "/processed_data";  
    FIR->processed_data_fd = open( temp_string.c_str(), O_RDONLY | O_NONBLOCK);

    //set FIFO capacity in bytes. 
    // NOTE: 1048576 bytes = 1 MiB is the largest allowed by the kernel
    set_FIFO_buffer_size( FIR->processed_data_fd, 1048576 );

    
    //	Use system call to launch the FIR engine, pass working_directory as command line parameter
    OS_commands = FIR_engine_cmdstr + " " + dir_name + " &"; 
    ret_val = run_OS_command( OS_commands );
    if ( ret_val != 0 ) {
      //an error was thrown when trying to launch the FIR engine. Set fatal error flag
      FIR->fatal_errors_encountered = true;
      error_file << "An error was thrown when trying to launch the FIR engine." << endl;
      return;
    }
  } //end activation steps that are prevented from occuring more than once            

} //end activate function. 




void run(LADSPA_Handle instance, unsigned long sample_count) {
  plugin_data_struct *FIR = (plugin_data_struct *)instance;
  LADSPA_Data *input = FIR->input;
  LADSPA_Data *output = FIR->output;

  FIR->call_counter++;


  if ( ! FIR->FIR_engine_is_ready ) {
    string temp_string;
    string working_directory(FIR->working_directory);
    //if this is the first call, calculate the update_interval and add it to FIFO messages
    if ( FIR->call_counter == 1 ) {
      unsigned long update_interval;
      //update_interval in microseconds = 1000000 * samples per frame / samples per second
      update_interval = 1000000 * sample_count / FIR->SR;
      //queue update_interval for transmission via FIFO
      string a_message = "update_interval = " + to_string(update_interval);
      string all_messages( FIR->messages );
      queue_message_for_FIFO( a_message, all_messages );
  		strcpy(FIR->messages, all_messages.c_str() );
    }
		if ( ! FIR->control_FIFO_open  ) {
		    //attempt to open the FIR engine control FIFO
        temp_string = working_directory + "/control";  
		    FIR->control_fifo_fd = open(temp_string.c_str(), O_WRONLY | O_NONBLOCK);
		    if (FIR->control_fifo_fd != -1) {
					FIR->control_FIFO_open = true;
				}
		}
    bool ready_signal;
    ready_signal = check_for_FIR_engine_ready_signal( FIR->processed_data_fd );
    if ( ready_signal ) {
      FIR->FIR_engine_is_ready = true;
      //the unprocessed data FIFO can now be opened as writer safely
      temp_string = working_directory + "/unprocessed_data";  
      FIR->unprocessed_data_fd = open(temp_string.c_str(), O_WRONLY | O_NONBLOCK);
//			error_file << "the FIR engine ready signal was received during call " << FIR->call_counter << endl;
    } 
  }

  // if the control FIFO is open and there are messages in the queue, attempt to send them
  if ( ( FIR->control_FIFO_open == true ) && ( strlen(FIR->messages) > 0 ) ) {
    string message_string( FIR->messages ); 
    write_to_control_FIFO(FIR->control_fifo_fd, message_string, FIR->control_FIFO_open );
    if ( message_string.size() == 0 ) FIR->messages[0] = '\0';
  } 

  //while the FIR engine is NOT ready...
  //  or if a fatal error has occured, 
  //  keep the plugin going by returning silence
  if ( ( FIR->fatal_errors_encountered ) || ( ! FIR->FIR_engine_is_ready ) ) {
    for (unsigned long index=0; index < sample_count; index++ ) {
      output[index] = 0.0;
    }
    return;
  }

  int ret_val; //used to hold the value returned by a FIFO read or write

  //write samples to processed_data FIFO
  ret_val = put_unprocessed_data ( FIR->unprocessed_data_fd, input, sample_count );

  //read samples from processed_data FIFO
  ret_val = get_processed_data ( FIR->processed_data_fd, output, sample_count );
  if ( ret_val < 1 ) {
		if ( ! FIR->is_startup ) {
			error_file << "Buffer Underrun Warning: processed_data was not available during call " << FIR->call_counter;
			error_file << "; " << sample_count << " zeros were returned to the host.";
			float latency = (float) sample_count;
			latency /= (float) FIR->SR;
			latency *= 1000.0;
			error_file << "   Latency increased by " << latency << " msec during call " << FIR->call_counter << endl;
      FIR->latent_sample_count += sample_count;
		}
		//RETURN ZEROS IF NO DATA IS AVAILABLE:
		for (unsigned long index=0; index < sample_count; index++ ) output[index] = 0.0;
	} else { 
    if ( FIR->latent_sample_count > 0 ) {
      //attempt to correct the latency by reading in FIR->latent_sample_count samples
      FIR->latent_sample_count -= sample_count;
      unsigned long num_compensated = sample_count;
      while ( ( FIR->latent_sample_count > 0 ) && ( ret_val > 0 ) ) {
        //read samples from processed_data FIFO
        ret_val = get_processed_data ( FIR->processed_data_fd, output, sample_count );
        if ( ret_val > 0 ) {
          FIR->latent_sample_count -= sample_count;
          num_compensated += sample_count;
        } 
      }
  		float latency = (float) num_compensated;
  		latency /= (float) FIR->SR;
  		latency *= 1000.0;
  		error_file << "Latency was decreased by " << latency << " msec during call " << FIR->call_counter << endl;
  		latency = (float) FIR->latent_sample_count;
  		latency /= (float) FIR->SR;
  		latency *= 1000.0;
      error_file << "   The latency is now " << latency << " msec." << endl; 
  		//RETURN ZEROS IF NO DATA IS AVAILABLE:
  		if ( ret_val < 1 ) for (unsigned long index=0; index < sample_count; index++ ) output[index] = 0.0;
    }
  } //done getting samples and correcting latency

	//check for end of startup conditions
  if ( ( FIR->is_startup ) && ( ret_val > 0 ) ) {
//		error_file << "Startup was completed during call " << FIR->call_counter << ".  Begin normal operation." << endl;
		FIR->is_startup = false;
	}
} //end run function 




void cleanup(LADSPA_Handle instance) {
  plugin_data_struct *FIR = (plugin_data_struct *)instance;

	//send the terminate message to the FIR engine
	string message = "action = terminate";
	bool is_open = FIR->control_FIFO_open;
	if ( is_open )
		write_to_control_FIFO ( FIR->control_fifo_fd, message, is_open );	

  //free memory used by LADSPA_Handle data structure
	free(instance);
} //end function cleanup




static class Initialiser {
public:
  Initialiser() {
    //open the error log file in append mode
    string temp_string;
    temp_string = "/dev/shm/LADSPA-FIR.log";  
  	error_file.open( temp_string.c_str(), ios_base::app );

    //LADSPA SPECIFIC INITIALIZATIONS:
    char **port_names;
    LADSPA_PortDescriptor *port_descriptors;
    LADSPA_PortRangeHint *port_range_hints;
    FIR_filter_Descriptor = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

    if (FIR_filter_Descriptor) {
      std::string text;
      //plugin descriptor info
      FIR_filter_Descriptor->UniqueID = 5225;
      FIR_filter_Descriptor->Label = "LADSPA_FIR";
      FIR_filter_Descriptor->Properties = LADSPA_PROPERTY_HARD_RT_CAPABLE;
      text = "LADSPA_FIR";
      FIR_filter_Descriptor->Name = strdup(text.c_str());
      FIR_filter_Descriptor->Maker = "Charlie Laub, 2020";
      FIR_filter_Descriptor->Copyright = "GPLv3";
      FIR_filter_Descriptor->PortCount = 7;

      //create storage for port_descriptors, port_range_hints, and port_names        
      port_descriptors = (LADSPA_PortDescriptor *)calloc(7,sizeof(LADSPA_PortDescriptor));
      FIR_filter_Descriptor->PortDescriptors = (const LADSPA_PortDescriptor *)port_descriptors;
      port_range_hints = (LADSPA_PortRangeHint *)calloc(7,sizeof(LADSPA_PortRangeHint));
      FIR_filter_Descriptor->PortRangeHints = (const LADSPA_PortRangeHint *)port_range_hints;
      port_names = (char **)calloc(7, sizeof(char*));
      FIR_filter_Descriptor->PortNames = (const char **)port_names;
      //done creating storage. now set the descriptor, range_hints, and name for each port:

      //port = FILTER_INDEX    
      port_descriptors[FILTER_INDEX] = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
      text = "filter_index";
      port_names[FILTER_INDEX] = strdup(text.c_str());
      port_range_hints[FILTER_INDEX].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_DEFAULT_0;
      port_range_hints[FILTER_INDEX].LowerBound = 0;

      //port = DATA_LENGTH    
      port_descriptors[DATA_LENGTH] = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
      text = "data_length";
      port_names[DATA_LENGTH] = strdup(text.c_str());
      port_range_hints[DATA_LENGTH].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_DEFAULT_0;
      port_range_hints[DATA_LENGTH].LowerBound = 0;

      //port = FFT_LENGTH    
      port_descriptors[FFT_LENGTH] = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
      text = "FFT_length";
      port_names[FFT_LENGTH] = strdup(text.c_str());
      port_range_hints[FFT_LENGTH].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_DEFAULT_0;
      port_range_hints[FFT_LENGTH].LowerBound = 0;

      //port = CYCLE_LATENCY    
      port_descriptors[CYCLE_LATENCY] = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
      text = "cycle_latency";
      port_names[CYCLE_LATENCY] = strdup(text.c_str());
      port_range_hints[CYCLE_LATENCY].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_DEFAULT_0;
      port_range_hints[CYCLE_LATENCY].LowerBound = 0;

      //port = FILTER_LABEL    
      port_descriptors[FILTER_LABEL] = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
      text = "filter_label";
      port_names[FILTER_LABEL] = strdup(text.c_str());
      port_range_hints[FILTER_LABEL].HintDescriptor = LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_DEFAULT_0;
      port_range_hints[FILTER_LABEL].LowerBound = 0;
      port_range_hints[FILTER_LABEL].UpperBound = 9999;

      //port = DATA_INPUT   
      port_descriptors[DATA_INPUT] = LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
      text = "Input";
      port_names[DATA_INPUT] = strdup(text.c_str());

      //port = DATA_OUTPUT
      port_descriptors[DATA_OUTPUT] = LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
      text = "Output";
      port_names[DATA_OUTPUT] = strdup(text.c_str());



      FIR_filter_Descriptor->activate = activate;
      FIR_filter_Descriptor->cleanup = cleanup;
      FIR_filter_Descriptor->connect_port = connectPort;
      FIR_filter_Descriptor->deactivate = NULL;
      FIR_filter_Descriptor->instantiate = instantiate;
      FIR_filter_Descriptor->run = run;
      FIR_filter_Descriptor->run_adding = NULL;
      FIR_filter_Descriptor->set_run_adding_gain = NULL;
    }
    //DONE WITH LADSPA-SPECIFIC INITIALIZATIONS
  } //end class initialization
  ~Initialiser() {
    if (FIR_filter_Descriptor) {
      free((LADSPA_PortDescriptor *)FIR_filter_Descriptor->PortDescriptors);
      free((char **)FIR_filter_Descriptor->PortNames);
      free((LADSPA_PortRangeHint *)FIR_filter_Descriptor->PortRangeHints);
      free(FIR_filter_Descriptor);
    }
  }                                      
} g_theInitialiser;

